"""Tracker rules processor mixin."""

from types import *
import time, whrandom

import Acquisition
from Persistence import Persistent
from DateTime import DateTime

DEBUG = 0

BIZDAYSTART = 9.0                       # 9:00 am
BIZDAYEND = 18.0                        # 6:00 pm

REGULAR_HOLIDAYS = {'Independence Day': '07/04',
                    'Christmas Day': '12/25',
                    'Thanksgiving': '11/25',
                    'New Years Day': '01/01',
                    'Easter': '04/04',
                    }

RULES_USER = '.rules.'

class TransactionAbort(Exception):
    pass

BadForm = "BadForm"           # String exception, needs to be caught elsewhere.

class Noodge(Acquisition.Implicit,
             Persistent,
             ):
    """Maintain and process the tracker's rule base.

    Rules are applied only to active issues - ie, those that are not
    in the completed, rejected, or deferred stage."""
    def __init__(self, rules=None):
        self.init_expr_tables()
        self.set_rules(rules)

    def set_rules(self, rules):
        self._rules = rules
    def init_expr_tables(self):
        # Map non-builtin conditions to method names.
        self._conditions = {'state': "state_condition",
                            'value': "issue_value",
                            }
        self._actions = {'notify': "notify_action",
                         }

    def execute_rules(self, issues=None):
        """Apply rules to all unclosed issues, or optional list 'issues'.

        Unclosed issues are pending, accepted, or deferred ones.

        Return value is a list of exceptions encountered - tuples of the
        exception, details, and rule - if any.

        NOTE BENE - Invokation without specifying the list of
        subject issues causes execute_rules to manage the transaction
        explicitly, begining and committing it within the context of each
        issue.  This means that any preceeding business in the transaction
        that calls execute_rules will be wiped clean.

        The state ticker for each issue is incremented after rules are
        executed for it, marking all surviving issue audited state as no
        longer volatile."""

        rules = self._rules
        if not rules:
            return
        if issues is not None:
            manage_transaction = 0
        else:
            catalog, cpaths = self._catalog, self._catalog.paths
            issues = map(lambda x, cpaths=cpaths: cpaths[x.data_record_id_],
                         catalog.searchResults({'stage': ['pending','deferred',
                                                          'accepted'],
                                                'type': ['tracker_issue']}))
            manage_transaction = 1
        todo = []

        exceptions = []

        for i in issues:
            transaction = None
            try:
                transaction = get_transaction()
                if manage_transaction:
                    transaction.begin()
            except:
                transaction = None

            aborted = 0
            try:
                #import pdb; pdb.set_trace()
                for rule in self._rules:
                    self.eval_rule(rule, i)
            except TransactionAbort, details:
                if transaction:
                    transaction.abort()
                aborted = 1
                exceptions.append(TransactionAbort, details, rule)
                if DEBUG >= 2:
                    Note("TransactionAbort(%s) encountered for %s\n",
                         details, `rule`)
            if not aborted and transaction and manage_transaction:
                transaction.commit()
            if DEBUG >= 1:
                i.audited_state.set('ranrules', i.tick(), RULES_USER)
            i.increment_tick()
                
        return exceptions

    def eval_rule(self, rule, issue):
        """Evaluate a rule, returning qualified actions.

        Returns 1 if any actions were executed, else 0 - unless:

        Raises TransactionAbort if an 'abort' action is executed.

        Raises BadForm if invalid rule is encountered."""

        # Rules are lists of data structures called contingencies:
        #
        #  [rule, ...] 
        # 
        # a rule contains 0 or more conditions and 1 or more
        # actions.
        #
        # <rule> :   ( ('condition', <condition>), <rule>, ... )
        #          | ('actions' (<1 or more actions>))
        #
        # A leading ('condition', <condition>), if any, must hold for any
        # of the subcontingencies to be evaluated.

        try:
            selector = rule[0]
        except (AttributeError, IndexError):
            return rule          # A literal.

        if selector == 'actions':
            return self.apply_actions(issue, rule[1:])
        elif type(selector) != TupleType:
            raise BadForm, ("Composite rule needs leading condition: %s"
                            % `rule`)
        elif selector[0] != 'condition':
            raise BadForm, ("Malformed Rule condition: %s"
                            % `rule`)
        else:
            did = 0
            if self.eval_condition(selector[1], issue):
                for i in rule[1:]:
                    did = self.eval_rule(i, issue) or did
            return did
        
    ###### Conditions ######

    def eval_condition(self, condition, issue):
        """Evaluate w.r.t issue and return truth or time (secs) val."""

        # condition is either a literal (number, etc - precisely, raises
        # AttributeError or IndexError upon index lookup) or a tuple
        # expression, either one of the following builtins or one defined
        # (according to its first-element selector) in the._conditions
        # dispatch table.
        # 
        #  ('and' <condition> ...)	  - logical conjunction
        #  ('or' <condition> ...)	  - logical disjunction
        #  ('not' <condition> ...)	  - logical negation
        #  ('>' <condition> <condition>)  - Greater than
        #  ('<' <condition> <condition>)  - Less than
        #  ('=' <condition> ...)	  - equals
        #  ('contains' <state-var> <value(s)>) - state var includes value
        #  ('volatile' <state-var> ...)	  - var got new value this tick

        if type(condition) == StringType:
            raise BadForm, ("Invalid condition - string: %s" % `condition`)

        try:
            cname = condition[0]
        except (AttributeError, IndexError, TypeError):
            return condition            # A literal.
        except TypeError, msg:
            raise TypeError, (msg, `condition`)

        if cname == 'and':
            for i in condition[1:]:
                if not self.eval_condition(i, issue):
                    return 0
            return 1
        elif cname == 'or':
            for i in condition[1:]:
                if self.eval_condition(i, issue):
                    return 1
            return 0
        elif cname == 'not':
            if len(condition) != 2:
                raise BadForm, ("Wrong number of args, need 1, got %d: %s"
                                % (len(condition)-1, `condition`))
            return not self.eval_condition(condition[1], issue)
        elif cname == '<':
            last = None
            for i in condition[1:]:
                next = self.eval_condition(i, issue)
                if last is not None and last >= next:
                    return 0
                last = next
            return 1
        elif cname == '>':
            last = None
            for i in condition[1:]:
                next = self.eval_condition(i, issue)
                if last is not None and last <= next:
                    return 0
                last = next
            return 1
        elif cname == '=':
            last = None
            for i in condition[1:]:
                next = self.eval_condition(i, issue)
                if last is None:
                    last = next
                elif last != next:
                    return 0
            return 1
        elif cname == 'contains':
            if len(condition) != 3:
                raise BadForm, ("Wrong number of args, need 2, got %d: %s"
                                % (len(condition)-1, `condition`))
            values = issue.audited_state.get(condition[1])
            if type(values) != DictType:
                values = {values: None}
            return values.has_key(condition[2])
        elif cname == 'volatile':
            history = issue.audited_state.history
            tick = issue.tick()
            for i in condition[1:]:
                if not history(i, tick=tick):
                    return 0
            return 1
        else:
            # Get from the table.
            methnm = self._conditions.get(cname, None)
            if not methnm:
                raise BadForm, ("Unrecognized condition '%s'"
                                % cname)
            elif not hasattr(self, methnm):
                raise SystemError, ("Improperly registered condition '%s'"
                                    % cname)
            meth = getattr(self, methnm)
            try:
                return apply(meth, (issue,) + tuple(condition[1:]))
            except (BadForm, TypeError), err:
                raise BadForm, str(err) + ": " + `condition`

    def state_condition(self, issue, var=None, value=None, priorcount=None):
        """Return time in decimal business days since state var was set to
        value, or 0 if not.
        
        Many forms:

         ('state') returns "age" of the issue - decimal days since creation.

         ('state' 'var') or ('state' 'var' None 0)
                   returns duration since var's most recent setting, or 0.

         ('state' 'var' val)
                   returns duration that varname has had current setting
                   equal to value, or 0 if it doesn't have that value.
                                 
         ('state' 'var' None n>0)
                   returns time since var got its n'th prior setting.

         ('state' 'var' val 0)
                   returns time since varname was last set to val, or 0.

         ('state' 'var' val n>0)
                   returns time since var got its n'th prior setting to
                   val, or 0.
        
        When value is a list, a value match for any of the items on the
        list satisfies."""

        state = issue.audited_state

        if not var:                # ('state')
            if value is not None:
                raise BadForm, "Invalid: value without state var"

            return state.age()

        x = ['unique']
        if state.get(var, x) is x:
            return 0                    # no setting for var

        # Assert: var has at least one setting.

        if value is None:
            if priorcount in [None, 0]:
                # ('state', "var") or ('state', "var", None, 0)
                return age(state.history(varname=var, max=1)[0][-1],
                                business=1)
            else:                       # ('state', "var", None, X):
                got = state.history(varname=var, max=priorcount+1)
                if len(got) < (priorcount+1):
                    return 0
                else:
                    return age(got[-1][-1], business=1)
        else:

            if type(value) != ListType:
                value = [value]

            if priorcount is None:      # ('state', "var", value):
                curval = state.get(var)
                if curval not in value:
                    return 0
                return age(state.history(varname=var, max=1)[0][-2],
                           business=1)

            elif priorcount == 0:       # ('state', "var", value, 0):
                got = state.history(varname=var, value=value, max=1)
                if not got:
                    return 0
                else:
                    return age(got[0][-2], business=1)

            elif priorcount > 0:        # ('state', "var", value, n):
                got = state.history(varname=var,
                                    value=value, max=1+priorcount)
                if len(got) < (priorcount):
                    return 0
                else:
                    return age(got[-1][-2], business=1)

            else:
                raise BadForm, "Invalid 'state' condition priorcount"

    def issue_value(self, issue, var):
        """Return value of issue variable; acquires tracker values, as well."""

        if not hasattr(issue, var):
            raise Attribute, "No such tracker variable '%s'" % var

        return getattr(issue, var)

    ###### Actions ######

    def apply_actions(self, issue, actions):
        """Invoke the specified actions in the context of the issue."""

        #  ('set' <state-var> <value>)      - Set state var to value
        #  ('include' <state-var> <value>)  - State var gets value among others
        #  ('remove' <state-var> <value>)   - State var no longer has value
        #  ('notify' <role> <subject> <msg>) - Send a notice, text is dtml
        #  ('abort' <msg>)		    - Forbidden - rollback transaction

        state = issue.audited_state
        for a in actions:
            try:
                selector = a[0]
            except (AttributeError, IndexError), err:
                raise BadForm, ("Invalid action - string: %s" % `a`)

            if selector == 'set':
                if len(a) != 3:
                    raise BadForm, ("Invalid 'set' action - "
                                    "wrong # of args: %s" % `a`)
                state.set(a[1], a[2], RULES_USER)
            elif selector == 'include':
                if len(a) != 3:
                    raise BadForm, ("Invalid 'include' action - "
                                    "wrong # of args: %s" % `a`)
                curval = state.get(a[1], {})
                if type(curval) is not DictType:
                    # Yes, we automatically promote to being a bag.
                    curval = {curval: None}
                newval = {}; newval.update(curval)
                newval[a[2]] = None
                if newval != curval:
                    state.set(a[1], newval, RULES_USER)
            elif selector == 'remove':
                if len(a) != 3:
                    raise BadForm, ("Invalid 'remove' action - "
                                    "wrong # of args: %s" % `a`)
                curval = state.get(a[1], {})
                if type(curval) is not DictType:
                    # Yes, we automatically promote to being a bag.
                    curval = {curval: None}
                newval = {}; newval.update(curval)
                del newval[a[2]]
                if newval != curval:
                    state.set(a[1], newval, RULES_USER)
            elif selector == 'abort':
                errargs = tuple(a[1:])
                if not errargs: errargs = None
                raise TransactionAbort, errargs
            else:
                # Get from the table.
                methnm = self._actions.get(a[0], None)
                if not methnm:
                    raise BadForm, ("Unrecognized action '%s'"
                                    % a[0])
                elif not hasattr(self, methnm):
                    raise SystemError, ("Improperly registered action '%s'"
                                        % a[0])
                meth = getattr(self, methnm)
                try:
                    apply(meth, (issue, RULES_USER) + tuple(a[1:]))
                except (BadForm, TypeError), err:
                    raise BadForm, str(err) + ": " + `a`
        
        return 1

    def notify_action(self, issue, user, who, subj, msg):
        # *** set 'msgfromrole' history var
        sender = self.adminaddr
        return self._notifier.notify(issue, who, subj, msg) or 1

def Note(text, *args):
    import sys
    try:
        msg = text % args
    except TypeError:
        msg = "Note - faulty note: %s, %s" % (`text`, `args`)
    __traceback_info__ = ("Note:", msg)
    try:
        sys.stderr.write(msg)
    except IOError:
        pass

def age(since, until=None, business=0, holidates=None,
        bizdaystart=BIZDAYSTART, bizdayend=BIZDAYEND):
    """Return age in decimal (float) number of days.

    Optional end time 'until', defaults to current moment.

    Optional flag 'business' means compute according to business time
    - excluding non-business hours, weekends, and specified holidays.

    Optional holidates is a list of DateTimes.Dates() format dates
    designating holidays to be excluded from business hours.
    \(Holidates falling on a weekend day are shifted to next
    non-exempted weekday before being exempted.)

    Optional bizdaystart is the start time of a business day, expressed
    *as a decimal number*: the integral hour, then the decimal portion
    thereof.

    Optional bizdayend is the companion day-end time for bizdaystart."""

    now = DateTime()
    if type(since) != type(now): since = DateTime(since)
    if not until: until = now
    elif type(until) != type(now): until = DateTime(until)

    if not business:
        return until - since

    # Business-hours computation onwards.

    # factor to normalize number of business hours/day:
    daylenfactor = 24/(bizdayend - bizdaystart)
    if daylenfactor < 0:
        raise 'DateTimeError', ('business day end (%s) is before start (%s)?? '
                                % (bizdayend, bizdaystart))

    # Shift the endpoints to nearest contained business days:
    since = nearest_bizday(since, forwards=1, holidates=holidates)
    until = nearest_bizday(until, forwards=0, holidates=holidates)
    if not since.lessThan(until):
        return 0
    # Nail down the amount of time in the head & tail days:
    head = tail = 0
    sincetod = since - since.earliestTime()
    if sincetod < (bizdaystart/24):
        # Increment the time to the start of the business day:
        since = since + ((bizdaystart/24) - sincetod)
        sincetod = since - since.earliestTime()
    untiltod = until - until.earliestTime()
    if untiltod > (bizdayend/24):
        # Decrement the time to the end of the business day:
        until = until - (untiltod - (bizdayend/24))
        untiltod = until - until.earliestTime()
    if not since.lessThan(until):
        return 0
    if since.Date() == until.Date():
        return (until - since) * daylenfactor
    else:
        subtotal = (((bizdayend/24 - sincetod) + (untiltod - bizdaystart/24))
                    * daylenfactor)

    # Ok, now we just need to deal with intervening days.
    since = since.latestTime()
    until = until.earliestTime()
    
    # Num of intervening days, sans head/tail
    interdays = int(until - since)

    if interdays == 0:
        return subtotal

    # Omit saturdays and sundays.
    dow_count = since.dow() + interdays
    weekend_days = dow_count/7 + (dow_count+1)/7
    interdays = interdays - weekend_days

    if holidates:
        # Omit intervening holidays.
        # Increment to the first non-weekend, non-considered day if necessary.
        exempted = []
        for htime in holidates:
            if since <= htime <= until:
                interdays = interdays - 1
                if interdays <= 0:
                    return subtotal

    return interdays + subtotal
    
def nearest_bizday(ztime, forwards, holidates):
    """Return the nearest business day to specified time.

    If we must shift forward, return the beginning of the soonest day.
    If backward, return the end of the most recent day.

    holidates are DateTime objects representing holiday dates to exclude."""

    # Each time around we concentrate on *either* the weekend *or* holiday...
    doingweekend = 0
    # ... so we can process the shifts consistently at the bottom.  Thus
    # we retraverse the loop after any shift to assess the newly found date.
    shift = 1
    # We only need to do the (relatively expensive) earliest/latest shifts
    # the first time - 'extremed' marks that. 
    extremed = 0
    while shift:
        shift = 0

        if not doingweekend:
            if ztime.dow() == 6:
                if forwards: shift = 2
                else: shift = -1
                doingweekend = 1
            elif ztime.dow() == 0:
                if forwards: shift = 1
                else: shift = -2
                doingweekend = 1
        else:
            doingweekend = 0

        if not doingweekend and holidates:
            thisDate = ztime.Date()
            for h in holidates:
                if thisDate == h.Date():
                    if forwards: shift = 1
                    else: shift = -1
        
        if shift:
            if shift > 0:
                if extremed == 0:
                    ztime = extremed = (ztime + shift).earliestTime()
                else: ztime = extremed = ztime + shift
            else:
                if extremed == 0:
                    ztime = extremed = (ztime + shift).latestTime()
                else: ztime = ztime + shift

    return ztime

class HoliDatesOff:
    """A cache for "regular" holidays, ie those determined by a month/day.

    '.add()' registers new month/day holidays.

    '.get()' takes a year and returns all the registered holiday dates, as
    DateTime objects, for that year - shifting any that fall on a weekend to
    the nearest subsequent non-holiday weekday.

    The HoliDatesOff class economizes on the DateTime object creation process
    by caching assortments for any encountered year.  The cache is invalidated
    any time a new date is entered."""
    def __init__(self, monthdaylist=[]):
        self._cache = {}                # {year: [dates]}
        self.monthdaylist = []
        self.add(monthdaylist)
    def add(self, dates):
        if type(dates) == type(""):
            dates = [dates]
        mdl = self.monthdaylist
        for i in dates:
            if i in mdl:
                continue
            mdl.append(i)
        self._cache = {}
    def get(self, year):
        year = str(year)
        if self._cache.has_key(year):
            print "from cache"
            return self._cache[year]
        got = []
        for i in self.monthdaylist:
            d = nearest_bizday(DateTime("%s/%s" % (year, i)), 1, got)
            got.append(DateTime(d.Date()))
        got = tuple(got)
        self._cache[year] = got
        return got

def getHolidayRegistry():
    return HoliDatesOff(REGULAR_HOLIDAYS.values())
