"""Issue tracking system implementation classes."""

__version__ = '$Revision: 1.24 $'

import time
import string
from types import *
from string import atoi

from Noodging import Noodge, BadForm, TransactionAbort, age


import Acquisition
import AccessControl.Role
import OFS.SimpleItem
from Globals import HTMLFile, default__class_init__
from Persistence import Persistent
from IOBTree import BTree
    

# Setting to 1 leave an entry in audit history for each rules run.
# Setting to 2 also produces debug flags on stderr.
DEBUG = 0

class TrackerBase(Noodge,
                  Acquisition.Implicit,
                  Persistent,
                  AccessControl.Role.RoleManager,
                  ):
    """Issue tracking system central implementation."""

    meta_class = 'TrackerIssue'

    isPrincipiaFolderish = 1

    __ac_permissions__ = (
        ('Browse',
         ('__getitem__', '__getattr__',
          'owners', 'keys', 'items', 'values', 'has_key',
          'update', 'get', 'length', '__len__', '__del__',
          '__delitem__', '__setitem__',
          ), ['Anonymous']),
        ('Create Issues', ('add_issue'), ['Anonymous']),
        ('Configure Tracker', (), ['TrackerOwner']),
        ('Supporter', (), ['TrackerOwner', 'Supporter']),
        )

    IMATRACKER = 1

    def __init__(self, id):
        __traceback_info__ = ("TrackerBase __init__ in", `id`)
        self.id = id
        self._notifier = Notifier()
        Noodge.__init__(self)
        self._issues = BTree()

    def __repr__(self):
        return ("<%s %s (%d issues) at 0x%s>"
                % (self.__class__.__name__, `self.id`, self.length(),
                   hex(id(self))[2:]))

    def add_issue(self, issue):
        """Add a new issue, and return the generated id."""
        if self._issues:
            id = self._issues.keys()[-1] + 1
        else:
            id = 1
        self._issues[id] = issue
        issue.id = str(id)
        return id

    def keys(self):
        """Return all issue keys, regardless of viewability of value."""
        return self._issues.keys()
    def items(self):
        """Return all keys and issues which we are privileged to view."""
        got = []
        for k, v in self._issues.items():
            got.append(k, v.__of__(self))
        return got
    def values(self):
        """Return all issues which we are privileged to view."""
        got = []
        for i in self.items():
            got.append(i[1])
        return got
    def has_key(self, id):
        """True if item exists for key, regardless of privilege to view it."""
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        return self._issues.has_key(id)
    def get(self, id, failobj=None):
        """Like generic sequence get, including failover."""
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        if self._issues.has_key(id):
            return self._issues[id]
        else:
            return failobj
    def length(self):
        """Number of all issues."""
        return self._issues.__len__()
    def __len__(self):
        """Number of all issues."""
        return self._issues.__len__()
##     def __del__(self):
##         """ """
##         for k in tuple(self.keys()):
##             del self._issues[k]
    def __delitem__(self, id):
        """ """
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        del self._issues[id]
    def __setitem__(self, id, issue):
        """ """
        __traceback_info__ = ("TrackerBase.__setitem__", `id`, `issue`)
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        self._issues[id] = issue

    def __getitem__(self, id):
        """ """
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        __traceback_info__ = ("TrackerBase.__getitem__", `id`)
        return self._issues[id].__of__(self)
    __getattr__ = __getitem__

    def owners(self):
        # Modelled on Extensions/ZopeSite.py:getOwners()
        got = []
        for k, v in self.__ac_local_roles__.items():
            if 'Owner' in v:
                got.append(k)
        return got

default__class_init__(TrackerBase)

class TrackerIssueBase(Acquisition.Implicit,
                       Persistent,
                       AccessControl.Role.RoleManager,
                       ):
    """Implement Issue state-change auditing, item-containment, permissions.

    See audited_state, add_state, register_state_changes for auditing features.

    The ZClass inheritor does the rest..."""

    # If we need to enable access to unprotected attributes.
    __allow_access_to_unprotected_attributes__ = 1

    isPrincipiaFolderish = 1

    IMATRACKERISSUE = 1

    meta_class = 'TrackerIssueBase'

    # XXX Permissions are going to have to be worked out!
    __ac_permissions__ = (
        ('Public Issue Interaction',
         ['add_item', '__getitem__', '__getattr__',
          'owners', 'keys', 'items', 'values',
          'has_key', 'update', 'get', 'length', 'length', '__len__',
          '__delitem__', '__setitem__', 'history', 'numid',
          'add_state', 'register_state_changes', 'statecond',
          'promotes',  # XXX XXX Expediency, permissions broke
          ], ['Anonymous']),
        ('Privileged Issue Interaction',
         [],                            # ['promotes'],
         ['Requester', 'IssueOwner', 'Supporter', 'TrackerOwner'],
         ),
        ('Tracker Staff',
         [],                            # ['promotes'],
         ['IssueOwner', 'Supporter', 'TrackerOwner'],
         ),
        ('Issue Overrides',
         [],
         ['IssueOwner', 'TrackerOwner'],
         )
        )

    def __init__(self, REQUEST=None):
        self.id = (REQUEST and REQUEST.get('id', '')) or ''
        self._messages = BTree()
        self._promoted = []
        self.audited_state = AuditedState()
        self.state_tick = 1

    def __repr__(self):
        return ("<%s %s (%d messages, %d promoted) at 0x%s>"
                % (self.__class__.__name__, `self.id`,
                   self.length(), len(self.promotes()),
                   hex(id(self))[2:]))

    def add_item(self, item, user):
        """Add a new item, returning generated id and item acquired."""
        items = self._messages
        if items:
            id = items.keys()[-1] + 1
        else:
            id = 1
        item.id = str(id)
        items[id] = item
        self.audited_state.set('newitem', id, user)
        return (id, item.__of__(self))

    def numid(self):
        """Return the numeric version of the id."""
        if type(self.id) == StringType:
            return atoi(self.id)
        else:
            return self.id

    def promotes(self, itemid=None, user=None, demote=None):
        """Return list of promoted items, or promote/demote one.

        Without an itemid, return the list of promoted items.  With just
        the itemid and user, add the itemid to the list - or remove it, if
        demote is set."""

        if itemid is None:
            return self._promoted[:]

        if type(itemid) == StringType:
            try:    itemid=atoi(itemid)
            except: raise AttributeError, itemid

        if demote is None:
            pubs = self._promoted
            if itemid not in pubs:
                pubs.append(itemid)
                self._promoted = pubs
            self.audited_state.set('promote', itemid, user)
        else:
            pubs = self._promoted
            if itemid in pubs:
                pubs.remove(itemid)
                self._promoted = pubs
            self.audited_state.set('demote', itemid, user)

    def execute_rules(self):
        """Invoke tracker rules on just this issue."""
        return self.aq_parent.execute_rules(issues=[self])

    def statecond(self, var=None, value=None, priorcount=None):
        """Expose Noodge.state_condition() method to dtml."""
        return self.state_condition(self, var=var, value=value,
                                    priorcount=priorcount)

    def tick(self):
        """Tick used to identify "volatile" settings of current transaction."""
        return self.state_tick

    def increment_tick(self):
        """Increment ticker by which audited state volitility is recognized."""
        self.state_tick = self.state_tick + 1

    def add_state(self, additions, REQUEST=None):
        """Institute the state settings that are to be audited."""
        # Create the audit trail:
        user = (REQUEST and `REQUEST.AUTHENTICATED_USER`) or ''
        setstate = self.audited_state.set
        for varname, value in additions:
            setstate(varname, value, user)

    def register_state_changes(self, REQUEST=None):
        """Check for tracked-attribute changes and register any found.

        Call this whenever the state of the issue has changed."""
        state = self.audited_state
        user = (REQUEST and `REQUEST.AUTHENTICATED_USER`) or ''
        for a, v in state.items():
            if not hasattr(self, a):
                # Event state vars, like 'newitem', are ephemeral...
                continue
            current = getattr(self, a)
            if current != v:
                state.set(a, current, user)
        return self.execute_rules()

    def history(self):
        """History of audited items."""
        return self.audited_state.history()

    def __getitem__(self, id):
        """ """
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        return self._messages[id].__of__(self)
    __getattr__ = __getitem__

    def keys(self, reverse=0):
        """ """
        got = list(self._messages.keys())
        if reverse: got.reverse()
        return got
    def items(self, reverse=0):
        """Return all keys and issues which we are privileged to view."""
        items = self._messages
        got = []
        for k, v in items.items():
            got.append(k, v.__of__(self))
        if reverse:
            got.reverse()
        return got
    def values(self, reverse=0):
        """Return all items which we are privileged to view."""
        got = []
        for i in self.items(reverse=reverse):
            got.append(i[1])
        return got
    def has_key(self, id):
        """True if item exists for key, regardless of privilege to view it."""
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        return self._messages.has_key(id)
    def get(self, id, failobj=None):
        """Like generic sequence get, including failover."""
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        items = self._messages
        if items.has_key(id):
            return items[id].__of__(self)
        else:
            return failobj
    def length(self):
        """Number of message items."""
        return self._messages.__len__()
    def __len__(self):
        """Number of message items."""
        return self._messages.__len__()

##     def __del__(self):
##         """ """
##         for k in tuple(self._messages.keys()):
##             del self._messages[k]
##         # XXX any other cleanup...

    def __delitem__(self, id):
        """ """
        # XXX is there any special stuff to do for ZClass deletion?
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        del self._messages[id]
    def __setitem__(self, id, issue):
        """ """
        __traceback_info__ = ("TrackerIssueBase.__setitem__", `id`, `issue`)
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        self._messages[id] = issue

    def __getitem__(self, id):
        """ """
        if type(id) == StringType:
            try:    id=atoi(id)
            except: raise AttributeError, id
        __traceback_info__ = ("TrackerIssueBase.__getitem__", `id`)
        return self._messages[id].__of__(self)

    def owners(self):
        lr = self.__ac_local_roles__ or {}
        got = []
        for k, v in lr.items():
            if 'TrackerOwner' in v:
                got.append(k)
        return got
default__class_init__(TrackerIssueBase)

class AuditedState(Acquisition.Explicit):

    # If we need to enable access to unprotected attributes.
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self):
        self._state = {}
        self._hist = []

    def set(self, varname, value, user):
        """Set current state and record old value in history."""
        self._state[varname] = value
        # self.tick() is got by acquisition.
        self._hist.insert(0, (varname, value, user, time.time(),
                              self.aq_parent.tick()))

    def get(self, varname, failobj=None):
        if not self._state.has_key(varname):
            return failobj
        return self._state[varname]

    def keys(self):
        return self._state.keys()
    def items(self):
        return self._state.items()

    def in_state(self, varname, value):
        return value == self.get(varname)
            
    def age(self):
        if not self._hist:
            return 0
        return age(self._hist[-1][-2], business=1)

    def history(self,
                varname=None, value=None, user=None, max=None, after=None,
                tick=None):
        """Return list of (name, val, oldval, user, time, tick) tuples.

        Order is more recent entries first.

        Optional parameters constrain the search.  varname, value, and user
        constrain entries to ones that contain that field.  (value and user
        can be lists of eligible candidates.) max specifies an upper limit
        to the number of entries returned.  after specifies a cutoff
        time.time(), entries logged before that time are excluded."""

        if varname != None and not self._state.has_key(varname):
            raise AttributeError, "Undefined varname '%s'" % varname
        got = []
        if varname == value == max == after == tick == None:
            __traceback_info__ = (`self`,)
            return self._hist[:]
        if value is not None and type(value) == StringType:
            user = [user]
        if user is not None and type(user) == StringType:
            user = [user]
        for n, v, u, t, i in self._hist:
            if after and t < after:
                break
            if varname and (n != varname):
                continue
            if value and (v not in value):
                continue
            if user and (u not in user):
                continue
            if tick:
                if i != tick:
                    return []
                return [(n, v, u, t, i)]
            got.append((n, v, u, t, i))
            if len(got) == max:
                break
        return got

class Notifier:
    def notify(self, issue, who, subj, msg):
        if DEBUG >= 2:
            Note("** Issue notify: %s, Notify: to: %s, subj:%s\nmsg: %s\n",
                 issue, who, subj, msg)

def sample():
    """Evolving test that returns a populated tracker."""
    tracker = TrackerBase('sample')
    tracker.abbrev = 'SAMP'
    issue = TrackerIssueBase()
    i = tracker.add_issue(issue)
    issue = tracker[i]
    audit = issue.audited_state
    user = 'test'
    class PseudoItem(Acquisition.Implicit):
        def __init__(self, title="<No title>", description="<No description>"):
            self.title = title
            self.description = description
            self.IMATRACKERITEM = 1
    item = PseudoItem("sample title", "sample description")
    issue.add_item(item, user)
    item = issue[1]
    additions = [('newitem', '0'),
                 ('stage', 'pending'),
                 ('punctuality', 'ok'),
                 ('correspondence', 'caught-up'),
                 ('priority', 'normal'),
                 ('urgency', 'normal'),
                 ('window', 'soon'),
                 ]
    class x:
        AUTHENTICATED_USER = 'test'
    issue.add_state(additions, x())
    return tracker, issue, item
