from DateTime import DateTime

BIZDAYSTART = 9.0                       # 9:00 am
BIZDAYEND = 18.0                        # 6:00 pm

def age(since, until=None, business=0, holidates=None,
        bizdaystart=BIZDAYSTART, bizdayend=BIZDAYEND):
    """Return age in decimal (float) number of days.

    Optional end time 'until', defaults to current moment.

    Optional flag 'business' means compute according to business time
    - excluding non-business hours, weekends, and specified holidays.

    Optional holidates is a list of DateTimes.Dates() format dates
    designating holidays to be excluded from business hours.
    \(Holidates falling on a weekend day are shifted to next
    non-exempted weekday before being exempted.)

    Optional bizdaystart is the start time of a business day, expressed
    *as a decimal number*: the integral hour, then the decimal portion
    thereof.

    Optional bizdayend is the companion day-end time for bizdaystart."""

    now = DateTime()
    if type(since) != type(now): since = DateTime(since)
    if not until: until = now
    elif type(until) != type(now): until = DateTime(until)

    if not business:
        return until - since

    # Business-hours computation onwards.

    # factor to normalize number of business hours/day:
    daylenfactor = 24/(bizdayend - bizdaystart)
    if daylenfactor < 0:
        raise 'DateTimeError', ('business day end (%s) is before start (%s)?? '
                                % (bizdayend, bizdaystart))

    # Shift the endpoints to nearest contained business days:
    since = nearest_bizday(since, forwards=1, holidates=holidates)
    until = nearest_bizday(until, forwards=0, holidates=holidates)
    if not since.lessThan(until):
        return 0
    # Nail down the amount of time in the head & tail days:
    head = tail = 0
    sincetod = since - since.earliestTime()
    if sincetod < (bizdaystart/24):
        # Increment the time to the start of the business day:
        since = since + ((bizdaystart/24) - sincetod)
        sincetod = since - since.earliestTime()
    untiltod = until - until.earliestTime()
    if untiltod > (bizdayend/24):
        # Decrement the time to the end of the business day:
        until = until - (untiltod - (bizdayend/24))
        untiltod = until - until.earliestTime()
    if not since.lessThan(until):
        return 0
    if since.Date() == until.Date():
        return (until - since) * daylenfactor
    else:
        subtotal = (((bizdayend/24 - sincetod) + (untiltod - bizdaystart/24))
                    * daylenfactor)

    # Ok, now we just need to deal with intervening days.
    since = since.latestTime()
    until = until.earliestTime()
    
    # Num of intervening days, sans head/tail
    interdays = int(until - since)

    if interdays == 0:
        return subtotal

    # Omit saturdays and sundays.
    dow_count = since.dow() + interdays
    weekend_days = dow_count/7 + (dow_count+1)/7
    interdays = interdays - weekend_days

    if holidates:
        # Omit intervening holidays.
        # Increment to the first non-weekend, non-considered day if necessary.
        exempted = []
        for htime in holidates:
            if since <= htime <= until:
                interdays = interdays - 1
                if interdays <= 0:
                    return subtotal

    return interdays + subtotal
    
def nearest_bizday(ztime, forwards, holidates):
    """Return the nearest business day to specified time.

    If we must shift forward, return the beginning of the soonest day.
    If backward, return the end of the most recent day.

    holidates are DateTime objects representing holiday dates to exclude."""

    # Each time around we concentrate on *either* the weekend *or* holiday...
    doingweekend = 0
    # ... so we can process the shifts consistently at the bottom.  Thus
    # we retraverse the loop after any shift to assess the newly found date.
    shift = 1
    # We only need to do the (relatively expensive) earliest/latest shifts
    # the first time - 'extremed' marks that. 
    extremed = 0
    while shift:
        shift = 0

        if not doingweekend:
            if ztime.dow() == 6:
                if forwards: shift = 2
                else: shift = -1
                doingweekend = 1
            elif ztime.dow() == 0:
                if forwards: shift = 1
                else: shift = -2
                doingweekend = 1
        else:
            doingweekend = 0

        if not doingweekend and holidates:
            thisDate = ztime.Date()
            for h in holidates:
                if thisDate == h.Date():
                    if forwards: shift = 1
                    else: shift = -1
        
        if shift:
            if shift > 0:
                if extremed == 0:
                    ztime = extremed = (ztime + shift).earliestTime()
                else: ztime = extremed = ztime + shift
            else:
                if extremed == 0:
                    ztime = extremed = (ztime + shift).latestTime()
                else: ztime = ztime + shift

    return ztime

class HoliDatesOff:
    """A cache for "regular" holidays, ie those determined by a month/day.

    '.add()' registers new month/day holidays.

    '.get()' takes a year and returns all the registered holiday dates, as
    DateTime objects, for that year - shifting any that fall on a weekend to
    the nearest subsequent non-holiday weekday.

    The HoliDatesOff class economizes on the DateTime object creation process
    by caching assortments for any encountered year.  The cache is invalidated
    any time a new date is entered."""
    def __init__(self, monthdaylist=[]):
        self._cache = {}                # {year: [dates]}
        self.monthdaylist = []
        self.add(monthdaylist)
    def add(self, dates):
        if type(dates) == type(""):
            dates = [dates]
        mdl = self.monthdaylist
        for i in dates:
            if i in mdl:
                continue
            mdl.append(i)
        self._cache = {}
    def get(self, year):
        year = str(year)
        if self._cache.has_key(year):
            print "from cache"
            return self._cache[year]
        got = []
        for i in self.monthdaylist:
            d = nearest_bizday(DateTime("%s/%s" % (year, i)), 1, got)
            got.append(DateTime(d.Date()))
        got = tuple(got)
        self._cache[year] = got
        return got

REGULAR_HOLIDAYS = {'Independence Day': '07/04',
                    'Christmas Day': '12/25',
                    'Thanksgiving': '11/25',
                    'New Years Day': '01/01',
                    'Easter': '04/04',
                    }

def getHolidayRegistry():
    return HoliDatesOff(REGULAR_HOLIDAYS.values())
